! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras

      use star_lib
      use star_def
      use const_def
	  use crlibm_lib
      
      implicit none
      
      ! these routines are called by the standard run_star check_model
      contains
      
	      subroutine extras_controls(s, ierr)
	         type (star_info), pointer :: s
	         integer, intent(out) :: ierr
	         ierr = 0
         
	         ! this is the place to set any procedure pointers you want to change
	         ! e.g., other_wind, other_mixing, other_energy  (see star_data.inc)
			 
			 s% other_D_mix => horizontal_intrusions
         
         
	      end subroutine extras_controls


				subroutine horizontal_intrusions(id, ierr)
					! other_D_mix hook works by just adding to s% D_mix
					integer, intent(in) :: id
					integer, intent(out) :: ierr
			 			 
					integer :: k, nz
					real(dp) :: kt,kmu,nu, R0, Pr, tau, r_th
			 			 
					type(star_info), pointer :: s
			 
						ierr = 0
			 
						! get the star_info pointer as s. 
						! for a full list of the info stored in the pointer, see star/public/star_data.inc
						call star_ptr(id,s,ierr)
						if(ierr/=0) stop 'failed getting star_ptr'
			 
						nz = 0
			 			 
						do k=1,s%nz
							! mixing types are defined in mlt/public/mlt_def.f
							if(s% gradL_composition_term(k) .lt. 0d0) then ! inverse mu gradient
								if(s% gradT(k) .lt. s% gradL(k)) then ! Ledoux stable
									nz = nz + 1
					 
									! get thermal, molecular, and viscous diffusivities from local quantities
									call get_diff_coeffs(s% rho(k),s% T(k),s% X(k),s% cp(k), &
										s% opacity(k), s% dominant_iso_for_thermohaline(k), kt, kmu, nu)
					 					 
									R0 = s% gradT_sub_grada(k) / s% gradL_composition_term(k) 
									! numerator is negative since Schw stable
									! denominator is negative since inverse mu gradient
									! never actually use this R0, since I assume maximum Nu_mu
					 
									Pr = nu/kt
									tau = kmu/kt
					 
									r_th = (r0-1d0)/(1d0/tau-1d0)
					 					 					 
									! numu is defined toward the end of this module
									! just assume maximum possible D_mix
									s% D_mix(k) = s% D_mix(k) + numu(1d0,0d0,Pr,tau) * kmu
					 
									s% mixing_type(k) = 4 ! flags as thermohaline mixing, will appear as pink in pgstar plots
								end if
							end if
						end do
			 
						write(*,*) 'mixing from intrusions in', nz, 'zones'
			 
					end subroutine horizontal_intrusions


		  
      
      
	      integer function extras_startup(s, id, restart, ierr)
	         type (star_info), pointer :: s
	         integer, intent(in) :: id
	         logical, intent(in) :: restart
	         integer, intent(out) :: ierr
	         ierr = 0
	         extras_startup = 0
	         if (.not. restart) then
	            call alloc_extra_info(s)
	         else ! it is a restart
	            call unpack_extra_info(s)
	         end if
	      end function extras_startup
      

	      ! returns either keep_going, retry, backup, or terminate.
	      integer function extras_check_model(s, id, id_extra)
	         type (star_info), pointer :: s
	         integer, intent(in) :: id, id_extra
	         extras_check_model = keep_going         
	         if (.false. .and. s% star_mass_h1 < 0.35d0) then
	            ! stop when star hydrogen mass drops to specified level
	            extras_check_model = terminate
	            write(*, *) 'have reached desired hydrogen mass'
	            return
	         end if


	         ! if you want to check multiple conditions, it can be useful
	         ! to set a different termination code depending on which
	         ! condition was triggered.  MESA provides 9 customizeable
	         ! termination codes, named t_xtra1 .. t_xtra9.  You can
	         ! customize the messages that will be printed upon exit by
	         ! setting the corresponding termination_code_str value.
	         ! termination_code_str(t_xtra1) = 'my termination condition'

	         ! by default, indicate where (in the code) MESA terminated
	         if (extras_check_model == terminate) s% termination_code = t_extras_check_model
	      end function extras_check_model


	      integer function how_many_extra_history_columns(s, id, id_extra)
	         type (star_info), pointer :: s
	         integer, intent(in) :: id, id_extra
	         how_many_extra_history_columns = 0
	      end function how_many_extra_history_columns
      
      
	      subroutine data_for_extra_history_columns(s, id, id_extra, n, names, vals, ierr)
	         type (star_info), pointer :: s
	         integer, intent(in) :: id, id_extra, n
	         character (len=maxlen_history_column_name) :: names(n)
	         real(dp) :: vals(n)
	         integer, intent(out) :: ierr
         
	         !note: do NOT add these names to history_columns.list
	         ! the history_columns.list is only for the built-in log column options.
	         ! it must not include the new column names you are adding here.
         
	         ierr = 0
	      end subroutine data_for_extra_history_columns

      
	      integer function how_many_extra_profile_columns(s, id, id_extra)
	         type (star_info), pointer :: s
	         integer, intent(in) :: id, id_extra
	         how_many_extra_profile_columns = 0
	      end function how_many_extra_profile_columns
      
      
	      subroutine data_for_extra_profile_columns(s, id, id_extra, n, nz, names, vals, ierr)
	         type (star_info), pointer :: s
	         integer, intent(in) :: id, id_extra, n, nz
	         character (len=maxlen_profile_column_name) :: names(n)
	         real(dp) :: vals(nz,n)
	         integer, intent(out) :: ierr
	         integer :: k
	         ierr = 0
         
	         !note: do NOT add these names to profile_columns.list
	         ! the profile_columns.list is only for the built-in profile column options.
	         ! it must not include the new column names you are adding here.

	         ! here is an example for adding a profile column
	         !if (n /= 1) stop 'data_for_extra_profile_columns'
	         !names(1) = 'beta'
	         !do k = 1, nz
	         !   vals(k,1) = s% Pgas(k)/s% P(k)
	         !end do
         
	      end subroutine data_for_extra_profile_columns
      

	      ! returns either keep_going or terminate.
	      ! note: cannot request retry or backup; extras_check_model can do that.
	      integer function extras_finish_step(s, id, id_extra)
	         type (star_info), pointer :: s
	         integer, intent(in) :: id, id_extra
	         integer :: ierr
	         extras_finish_step = keep_going
	         call store_extra_info(s)

	         ! to save a profile, 
	            ! s% need_to_save_profiles_now = .true.
	         ! to update the star log,
	            ! s% need_to_update_history_now = .true.

	         ! see extras_check_model for information about custom termination codes
	         ! by default, indicate where (in the code) MESA terminated
	         if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step
	      end function extras_finish_step
      
      
	      subroutine extras_after_evolve(s, id, id_extra, ierr)
	         type (star_info), pointer :: s
	         integer, intent(in) :: id, id_extra
	         integer, intent(out) :: ierr
	         ierr = 0
	      end subroutine extras_after_evolve
      
      
	      ! routines for saving and restoring extra data so can do restarts
         
	         ! put these defs at the top and delete from the following routines
	         !integer, parameter :: extra_info_alloc = 1
	         !integer, parameter :: extra_info_get = 2
	         !integer, parameter :: extra_info_put = 3
      
      
	      subroutine alloc_extra_info(s)
	         integer, parameter :: extra_info_alloc = 1
	         type (star_info), pointer :: s
	         call move_extra_info(s,extra_info_alloc)
	      end subroutine alloc_extra_info
      
      
	      subroutine unpack_extra_info(s)
	         integer, parameter :: extra_info_get = 2
	         type (star_info), pointer :: s
	         call move_extra_info(s,extra_info_get)
	      end subroutine unpack_extra_info
      
      
	      subroutine store_extra_info(s)
	         integer, parameter :: extra_info_put = 3
	         type (star_info), pointer :: s
	         call move_extra_info(s,extra_info_put)
	      end subroutine store_extra_info
      
      
	      subroutine move_extra_info(s,op)
	         integer, parameter :: extra_info_alloc = 1
	         integer, parameter :: extra_info_get = 2
	         integer, parameter :: extra_info_put = 3
	         type (star_info), pointer :: s
	         integer, intent(in) :: op
         
	         integer :: i, j, num_ints, num_dbls, ierr
         
	         i = 0
	         ! call move_int or move_flg    
	         num_ints = i
         
	         i = 0
	         ! call move_dbl       
         
	         num_dbls = i
         
	         if (op /= extra_info_alloc) return
	         if (num_ints == 0 .and. num_dbls == 0) return
         
	         ierr = 0
	         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
	         if (ierr /= 0) then
	            write(*,*) 'failed in star_alloc_extras'
	            write(*,*) 'alloc_extras num_ints', num_ints
	            write(*,*) 'alloc_extras num_dbls', num_dbls
	            stop 1
	         end if
         
	         contains
         
	         subroutine move_dbl(dbl)
	            real(dp) :: dbl
	            i = i+1
	            select case (op)
	            case (extra_info_get)
	               dbl = s% extra_work(i)
	            case (extra_info_put)
	               s% extra_work(i) = dbl
	            end select
	         end subroutine move_dbl
         
	         subroutine move_int(int)
	            integer :: int
	            i = i+1
	            select case (op)
	            case (extra_info_get)
	               int = s% extra_iwork(i)
	            case (extra_info_put)
	               s% extra_iwork(i) = int
	            end select
	         end subroutine move_int
         
	         subroutine move_flg(flg)
	            logical :: flg
	            i = i+1
	            select case (op)
	            case (extra_info_get)
	               flg = (s% extra_iwork(i) /= 0)
	            case (extra_info_put)
	               if (flg) then
	                  s% extra_iwork(i) = 1
	               else
	                  s% extra_iwork(i) = 0
	               end if
	            end select
	         end subroutine move_flg
      
	      end subroutine move_extra_info
		  
		 
! the following routines were all copied and slightly adapted (by chris)
! from mlt/private/mlt.f, revision 6791.	  

          double precision function numu(R0,r_th,prandtl,diffratio)
          !Function calculates Nu_mu from input parameters, following Brown et al. 2013.
          !Written by P. Garaud (2013). Please email pgaraud@ucsc.edu for troubleshooting. 

             real(dp), intent(in) :: R0,r_th,prandtl,diffratio
             real(dp) :: maxl2,maxl,lambdamax
             real(dp) :: myvars(2)
             integer :: ierr, iter

             ! Initialize guess using estimates from Brown et al. 2013
             call analytical_estimate_th(maxl,lambdamax,r_th,prandtl,diffratio)
                  
             myvars(1) = maxl
             myvars(2) = lambdamax

            !Call Newton relaxation algorithm
            call NR(myvars,prandtl,diffratio,R0,ierr)
          
            !If the growth rate is negative, then try another set of parameters as first guess.  
            !Repeat as many times as necessary until convergence is obtained.
            iter = 1
            do while((myvars(2)<0).or.(ierr /= 0)) 
               !write(*,*) 'Alternative', r_th,prandtl,diffratio,iter
            !Reset guess values
               myvars(1) = maxl
               myvars(2) = lambdamax
            !Call relaxation for slightly different Pr, tau, R0.
               call NR(myvars,prandtl*(1.+iter*1.d-2),diffratio,R0/(1.+iter*1.d-2),ierr)
            !If it converged this time, call NR for the real parameters.
               if(ierr.eq.0) call NR(myvars,prandtl,diffratio,R0,ierr)
               !write(*,*) prandtl,diffratio,R0,myvars(1),myvars(2),ierr
               !Otherwise, increase counter and try again.
               iter = iter + 1            
            enddo

            !Plug solution into "l^2" and lambda.
            maxl2 = myvars(1)*myvars(1)
            lambdamax = myvars(2) 
            !write(*,*) prandtl,diffratio,r_th,maxl2,lambdamax

            !Calculate Nu_mu using Formula (33) from Brown et al, with C = 7.
            numu = 1. + 49.*lambdamax*lambdamax/(diffratio*maxl2*(lambdamax+diffratio*maxl2))

          return
          end function numu 

          subroutine thermohaline_rhs(myx,myf,myj,prandtl,diffratio,R0)
          ! This routine is needed for the NR solver.
          ! Inputs the two following equations for lambda and maxl2:
          ! lambda^3 + a_2 lambda^2 + a_1 lambda + a_0 = 0 (eq. 19 of Brown et al.)
          ! b_2 lambda^2 + b_1 lambda + b_0 = 0 (eq. 20 of Brown et al.)
          ! Inputs f, the equations, and j, their jacobian.
          ! Written by P. Garaud (2013). Please email pgaraud@ucsc.edu for troubleshooting. 

          real(dp), intent(in) :: myx(2),  prandtl, diffratio, R0
          real(dp), intent(out) :: myf(2), myj(2,2)
          real(dp) :: a_2,a_1,a_0,b_2,b_1,b_0,myterm,myx1_2,myx1_3,myx1_4
 
          !This inputs the coefficients.
          b_2 = 1.+prandtl+diffratio
          myx1_2 = myx(1)*myx(1)
          myx1_3 = myx1_2*myx(1)
          myx1_4 = myx1_3*myx(1)
          a_2 = myx1_2*b_2
          myterm = diffratio*prandtl+prandtl+diffratio
          b_1 = 2*myx1_2*myterm
          a_1 = myx1_4*myterm + prandtl*(1. - (1./R0))
          b_0 = 3.*myx1_4*diffratio*prandtl + prandtl*(diffratio - (1./R0))
          a_0 = myx1_4*myx1_2*diffratio*prandtl + myx1_2*prandtl*(diffratio - (1./R0))

 !         write(*,*) a_2,a_1,a_0,b_2,b_1,b_0

          !These are equations 19 and 20
          myf(1) = ((myx(2) + a_2)*myx(2) + a_1)*myx(2) + a_0
          myf(2) = b_2*myx(2)*myx(2) + b_1*myx(2) + b_0
         
          !These are their Jacobians for the NR relaxation.
          myj(1,1) = 2*myx(1)*b_2*myx(2)*myx(2) + &
             4*myx1_3*myterm*myx(2) + 6*myx1_4*myx(1)*diffratio*prandtl   &
               + 2*myx(1)*prandtl*(diffratio - (1./R0))
          myj(1,2) = 3*myx(2)*myx(2) + 2*a_2*myx(2) + a_1
          myj(2,1) = 4*myx(1)*myterm*myx(2) + 12.*myx1_3*diffratio*prandtl
          myj(2,2) = 2*b_2*myx(2) + b_1
 
          return
          end subroutine thermohaline_rhs               

          subroutine analytical_estimate_th(maxl,lambdamax,r_th,prandtl,diffratio)
          !Inputs analytical estimates for l and lambda from Brown et al. 2013.

          real(dp) :: prandtl, diffratio, maxl, lambdamax, r_th, phi, maxl4, maxl6
         
          phi = diffratio/prandtl

          if(r_th.lt.min(0.5*phi/(1.+phi),0.5*(1.+phi)*(1.+phi)*(1.+phi)/(4.*phi))) then
             if(r_th.gt.prandtl) then
                maxl = pow_cr((1./(1.+phi)) - 2.*dsqrt(r_th*phi)/pow_cr(1.+phi,2.5d0),0.25d0)   ! Equation (B14)
                maxl4 = maxl*maxl*maxl*maxl
                maxl6 = maxl4*maxl*maxl
                lambdamax = 2*prandtl*phi*maxl6/(1.-(1.+phi)*maxl4)    ! Equation (B11)
             else
                maxl = dsqrt(dsqrt(1./(1.+phi)) - dsqrt(prandtl)*(1.+phi/((1.+phi)*(1.+phi))))  ! Equation (B5)
                lambdamax = dsqrt(prandtl) - prandtl*dsqrt(1.+phi)   !Equation (B5)
             endif
          else 
             maxl = pow_cr((1./3.)*(1.-r_th) + (1-r_th)*(1-r_th)*(5-4*phi)/27.,0.25d0)
                ! Equation (B19) carried to next order (doesn't work well otherwise)
             maxl4 = maxl*maxl*maxl*maxl
             maxl6 = maxl4*maxl*maxl
             lambdamax = 2*prandtl*phi*maxl6/(1.-(1.+phi)*maxl4) ! Equation (B11)
          endif
          if(lambdamax<0) then   ! shouldn't be needed, but just as precaution
             maxl = 0.5d0
             lambdamax = 0.5d0
          endif
         
          return
          end subroutine analytical_estimate_th

          subroutine NR(xrk,prandtl,diffratio,R0,ierr)
          ! Newton Relaxation routine used to solve cubic & quadratic in thermohaline case.
          ! Written by P. Garaud (2013). Please email pgaraud@ucsc.edu for troubleshooting. 

          real(dp), parameter :: acy = 1.d-13 ! accuracy of NR solution.
          integer, parameter :: niter = 20  ! max number of iterations allowed before giving up.
          integer, parameter :: &  !array dimension input parameters for dgesvx
                n = 2, &
                nrhs = 1, &
                lda = n, &
                ldaf = n, &
                ldb = n, &
                ldx = n

          integer :: iter,ierr
          real(dp) :: xrk(2), f(2) ! Functions f 
          real(dp) :: j(2,2) ! Jacobian
          real(dp) :: err,errold ! Error at each iteration
          real(dp) :: x1_sav,x2_sav
          real(dp) :: prandtl, diffratio, R0
          real(dp) :: A(lda,n), AF(ldaf,n), R(n), C(n), B(ldb,nrhs), X(ldx,nrhs), &
                rcond, ferr(nrhs), berr(nrhs), work(4*n)
          character :: fact, trans, equed
          integer :: ipiv(n), iwork(n)

          include 'formats'

          !Initialize flags and other counters.
          ierr = 0
          iter = 0
          err = 0d0
          errold = 0d0
          !Save input guess (probably not necessary here but useful in other routines)
          x1_sav = xrk(1)
          x2_sav = xrk(2)

          !While error is too large .and. decreasing, iterate.
          do while ((err.gt.acy).and.(ierr.eq.0).and.(iter.lt.niter))
             call thermohaline_rhs(xrk,f,j,prandtl,diffratio,R0)    
            
             fact = 'E'
             trans = 'N'
             equed = ''
               
             A  = j
             B(1,1) = f(1)
             B(2,1) = f(2)
             call dgesvx( fact, trans, n, nrhs, A, lda, AF, ldaf, ipiv, &
                equed, r, c, B, ldb, x, ldx, rcond, ferr, berr, &
                work, iwork, ierr )
 
             if (ierr /= 0) then
                !write(*,*) 'dgesvx failed in thermohaline routine', iter
                !write(*,2) j(1,1),j(1,2)
                !write(*,2) j(2,1),j(2,2)
             else
                iter = iter + 1
                f(1) = X(1,1)
                f(2) = X(2,1)
                err = dsqrt(f(1)*f(1)+f(2)*f(2)) ! Calculate the new error
                ! If, after a while, the error is still not decreasing, give up and exit NR.
                ! Otherwise, continue.
                if((iter.gt.5).and.(err.gt.errold)) then              
                   ! Write(*,2) 'Error not decreasing at iter', iter, err, errold
                   ierr = 1
                   ! Reset xs and exit loop.
                   xrk(1) = x1_sav
                   xrk(2) = x2_sav                   
                else
                   xrk = xrk - f ! The solution is now in f, so update x 
                   errold = err
                endif
             endif
          enddo
            
          !if(iter.eq.niter) write(*,2) 'Failed to converge'
          return
          end subroutine NR

          subroutine get_diff_coeffs(rho,T,X,cp,opacity,dominant_iso_for_thermohaline,&
			  kt,kmu,vis)
			  ! stolen from mlt/private/mlt.f.  
			  ! inputs: rho, T, X, opacity,dominant_iso_for_thermohaline
			  ! outputs: nu, kmu, and kt

          use chem_def, only: chem_isos
          use crlibm_lib

		  real(dp), intent(in) :: rho,T,X,cp,opacity
		  integer, intent(in) :: dominant_iso_for_thermohaline
          real(dp), intent(out) :: kt,kmu,vis
		  real(dp) :: K
		  real(dp) :: qe4
          real(dp) :: loglambdah,loglambdacx,loglambdacy,ccx,ccy
		  real(dp) :: nu_rad, nu_mol
          real(dp) :: Bcoeff
          real(dp) :: chemA,chemZ,acx,acy       
          real(dp), parameter :: sqrt5 = sqrt(5d0)
                 
          K = 4*crad*clight*T*T*T/(3*opacity*rho) ! thermal conductivity
				 
          kt = K/(Cp*rho)       ! thermal diffusivity (assumes radiatively dominated)
          qe4=qe*qe*qe*qe
         
          ! Log Lambda for pure H (equation 10 from Proffitt Michaud 93)
          loglambdah = -19.26 - 0.5*log_cr(rho) + 1.5*log_cr(T) - 0.5*log_cr(1. + 0.5*(1+X)) 
          nu_rad = 4*crad*T*T*T*T/(15*clight*opacity*rho*rho) ! radiative viscosity
          nu_mol = 0.406*sqrt(amu)*pow_cr(boltzm*T,2.5d0)/(qe4*loglambdah*rho) 
          ! From Spitzer "Physics of Fully Ionized Gases equation 5-54
          ! Assumes pure H. Still trying to work out what it would be for a mixture. 
          vis = nu_mol + nu_rad   ! total viscosity

          ! The following is from Proffitt & Michaud, 1993.
          ! Their constant B (equation 15)
          Bcoeff = (15./16.)*sqrt(2.*amu/(5*pi))*pow_cr(boltzm,2.5d0)/qe4
          ! Extract what species drives the thermohaline convection
          chemA = chem_isos%Z_plus_N(dominant_iso_for_thermohaline)
          chemZ = chem_isos%Z(dominant_iso_for_thermohaline)

          if(chemZ.gt.2) then
          ! This is if the driving chemical is NOT He.
             ! Log Lambda for H-dominant chem mixture (equation 10)
             loglambdacx = loglambdah - log_cr(chemz)  
             ! Log Lambda for He-dominant chem mixture (equation 10)
             loglambdacy = loglambdah - log_cr(2.*chemz)
             ! Calculation of C_ij coeffs (equation 12)
             ccx = log_cr(exp_cr(1.2*loglambdacx)+1.)/1.2
             ccy = log_cr(exp_cr(1.2*loglambdacy)+1.)/1.2
             ! Reduced masses (I had to guess, from Bahcall & Loeb 1990), with H and He
             acx = (1.*chemA)/(1.+chemA)
             acy = 4*chemA/(4.+chemA)
             ! My formula (see notes) based on Proffitt and Michaud 1993
             kmu = 2*Bcoeff*pow_cr(T,2.5d0)/(sqrt5*rho*chemZ*chemZ)/ &
                (X*sqrt(acx)*ccx + (1-X)*sqrt(acy)*ccy)

          else
             ! Log Lambda for H-He mixture (equation 10)
             loglambdah = -19.26 - log_cr(2.d0) - 0.5*log_cr(rho) + &
                1.5*log_cr(T) - 0.5*log_cr(1. + 0.5*(1+X)) 
             ! Calculation of C_ij coeffs (equation 12)
             ccy = log_cr(exp_cr(1.2*loglambdah)+1.)/1.2
             ! My formula (see notes) based on Proffitt and Michaud 1993
             kmu = (Bcoeff*pow_cr(T,2.5d0)/(rho*ccy))*(3+X)/((1+X)*(3+5*X)*(0.7+0.3*X))
            
          endif
          ! write(57,*) kt,kmu,vis,chemZ

          end subroutine get_diff_coeffs

		  
      end module run_star_extras
      
